/* HTTP Proxy Server, by Mark Kamichoff [kamicm@rpi.edu] *
 * Original Filename: doserver.c                         */

int doserver(struct hostent *hptr)	//
{
	struct hostent *chptr;		// Address structure about remote server
	struct in_addr **pptr;		// Address list from remote server
	struct sockaddr_in raddr;	// Remote address structure

	char sendbuffer[1280];		/* used when sending url and/or content length */
	char buffer[1280];		/* our max GET line can be 1000 chars */
	char buffer2[1024];		/* this is for all other headers */
	char host[1280];		// Remote hostname
	char portnum[10];		// Report port number
	char path_ver[1280];
	char content_length[10];
	int flag;			/* whether we have crlfcrlf or not */

	char method[8];
	char method_num;		/* 0=GET, 1=HEAD, 2=POST */
	int n;				/* number of character read */
	int i, j, k;			/* counter */

	bzero(&raddr, sizeof(raddr));	// Sets the raddr structure to 0
	bzero(&buffer, sizeof(buffer));		// same for these too
	bzero(&buffer2, sizeof(buffer2));
	bzero(&sendbuffer, sizeof(sendbuffer));

//	printf("Connection from: %s\n",hptr->h_name);

	// Reads the first line, and ends when there is a \r, or an \n for buggy clients
	for(i=0; i<1280; ++i)
	{
		if((buffer[i] == '\r') || (buffer[i] == '\n'))
			break;
		if((n = read(clientfd, &buffer[i], 1)) != 1)
			return -1;
		if((buffer[i] == '\r') || (buffer[i] == '\n'))
			break;
	}

	if(i == 1280)
	{
		fprintf(stderr,"Error: Request line too large (1280 bytes)\n");
		if(write(clientfd,"HTTP/1.0 400 Bad Request\r\nServer: Netprog Proxy Server",54) != 54)
			fprintf(stderr,"Error: additionally, a write error was encountered while delivering the HTTP 501 error...\n");
		return -1;
	}

	read(clientfd, &buffer[i], 1);	// Sucks up the \n since we have \r\n usually

	buffer[i] = 0;			// Add the null terminator, since we use C-style string functions

//	printf("Parsed request: %s\n", buffer);
	printf("%s: %s\n", hptr->h_name, buffer);

	for(i=0; (buffer[i] != ' ') && (i<8); i++)	// Extracts the method
		method[i] = buffer[i];
	method[i] = 0;

	// This determines what method we just parsed, dies if not GET,HEAD,POST
	if(strcmp(method,"GET")==0)		// GET
		method_num=0;
	else if(strcmp(method,"HEAD")==0)	// HEAD
		method_num=1;
	else if(strcmp(method,"POST")==0)	// POST
		method_num=2;
	else
	{
		fprintf(stderr,"Error: bad request, not get/head/post, outputting HTTP 501\n");
		if(write(clientfd,"HTTP/1.0 501 Not Supported\r\nServer: Netprog Proxy Server",56) != 56)
			fprintf(stderr,"Error: additionally, a write error was encountered while delivering the HTTP 501 error...\n");
		return -1;
	}
//	printf("Debug: Method was %s\n",method);

	/* we set k to 11 or 12, whether the method is 3 chars or 4 chars respectively */
	j=0;	/* to account for both cases */

	if(method_num == 0)		// GET
		k = 11;
	else if(method_num == 1)	// HEAD
		k = 12;
	else				// POST
		k = 12;

	if(buffer[k-7] != 'h')
	{
		fprintf(stderr,"Error: Client sent a request with a non-http protocol (or bad syntax)\n");
		if(write(clientfd,"HTTP/1.0 400 Bad Request\r\nServer: Netprog Proxy Server",54) != 54)
			fprintf(stderr,"Error: additionally, a write error was encountered while delivering the HTTP 501 error...\n");
		return -1;
	}

	for(i=k; (buffer[i] != '/') && (buffer[i] != ':') && (i<1280); ++i)
	{
		host[i-k] = buffer[i];
	}
	host[i-k] = 0;	/* null */

	if(strlen(host) > 1000)		// We won't deal with things this long
	{
		fprintf(stderr, "Error: client %s sent hostname > 1000 chars, possible DoS ... \n", hptr->h_name);
		return -1;
	}	

	// Check if we have a custom port number
	if(buffer[i] == ':')	// We have a port number other than 80
	{
		for(j=1; (buffer[i+j] != '/') && (j < 6); j++)
		{
			portnum[j-1] = buffer[i+j];
		}
		portnum[j-1] =0;
		if((atoi(portnum) > 0) && (atoi(portnum) < 65537))
		{
			printf("Custom port %d\n",atoi(portnum));
		}
		else
		{
			fprintf(stderr,"Error: client %s requested invalid portnum of %s, dumping\n",hptr->h_name,portnum);
			return -1;
		}
	}
	else			// They didn't specify a port number, defaulting to 80
	{
		strcpy(portnum,"80");
	}

	// This parses the /index.html HTTP/1.1 part out of the buffer, which is eventually sent to the client
	for(k=(i+j);(k<1280) && (k<strlen(buffer));k++)
	{
		path_ver[k-(i+j)] = buffer[k];
	}
	path_ver[k-(i+j)]=0;

//	printf("DEBUG IS: %s \n",path_ver);
//	printf("We got: %s as request and hostname is: %s\n", buffer, host);

	/* Open the new socket for talking to the remote HTTP server */
	if((requestfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		fprintf(stderr, "Error: Socket could not be opened, dumping client...");
		return -1;
	}

	if((chptr = gethostbyname(host)) == NULL)	// Resolves the hostname we parsed
	{
		fprintf(stderr, "Error: Could not resolve hostname %s, outputting HTML error\n", &host);

		sleep(1);	/* need this so the client finishes sending headers and \r\n\r\n */

		snprintf(buffer2,sizeof(buffer2),
			"HTTP/1.0 400 Bad Request\r\n\r\n<html><head><title>400 Error
			</title></head><body bgcolor=black text=white link=white alink=white vlink=white>
			<h2>Proxy Error: Could not resolve %s</h2><hr><i>
			Proxilium proxy server 0.1-pre1-netprog, <a href=http://www.prolixium.com/>
			www.prolixium.com</a></i></body></html>",
			&host);

		if(write(clientfd,buffer2,strlen(buffer2)) < 1)
		{
			fprintf(stderr,"Error: Writing an error message to client %s which was an error trying to resolve %s!  Ha!\n",hptr->h_name, &host);
			return -1;
		}
		return -1;
	}

	/* copies memory from the first ip address in the address list
	   to the address of raddr (remote web server we wanna talk to) */

	pptr = (struct in_addr **) chptr->h_addr_list;
	memcpy(&raddr.sin_addr, *pptr, sizeof(struct in_addr));
	raddr.sin_port = htons(atoi(portnum));	/* arragh, we need port! */
	raddr.sin_family = AF_INET;

//	printf("Info: %s resolved to %s, now connecting to port %d...\n",
//		&host,
//		inet_ntoa(raddr.sin_addr),
//		ntohs(raddr.sin_port));

	if(connect(requestfd, (SA *) &raddr, sizeof(raddr)) < 0)
	{
		fprintf(stderr,"Error: Can't connect to %s, sending HTMLified error message\n", &host);

		sleep(1);	/* need this so the client finishes sending headers and \r\n\r\n */

		snprintf(buffer2,sizeof(buffer2),
			"HTTP/1.1 400 Bad Request\r\n\r\n<html><head><title>400 Error
			</title></head><body bgcolor=black text=white link=white alink=white vlink=white>
			<h2>Proxy Error: Could not connect to %s:%s</h2><hr><i>
			Proxilium proxy server 0.1-pre1-netprog, <a href=http://www.prolixium.com/>
			www.prolixium.com</a></i></body></html>",
			inet_ntoa(raddr.sin_addr), portnum);

		if(write(clientfd,buffer2,strlen(buffer2)) < 1)
		{
			fprintf(stderr,"Error: Writing an error message to client %s which was an error trying to connect to %s!  Ha!\n",hptr->h_name,&host);
			return -1;
		}

		return -1;
	}

	if(method_num == 0)			// get
		strcpy(sendbuffer,"GET ");
	else if(method_num == 1)		// head
		strcpy(sendbuffer,"HEAD ");
	else					// post
		strcpy(sendbuffer,"POST ");

	// Add on the /path/to/index.html HTTP/1.0\r\n string
	strcat(sendbuffer,path_ver);
	sendbuffer[strlen(sendbuffer)-1] = '0';		// Change 1.1 to 1.0
	strcat(sendbuffer,"\r\n");
	

	if(write(requestfd,&sendbuffer,strlen(sendbuffer)) != strlen(sendbuffer))
	{
		fprintf(stderr,"Error: HTTP directive size mismatch, dumping client\n");
		return -1;
	}

	/* Zeros the buffer2 */
	bzero(buffer2,sizeof(buffer2));

	/* this reads headers, parses them for \r\n\r\n, and passes them to the server */
	flag=1;		// We figure we have one to begin with, from the METHOD
	while(flag != 3)
	{
		for(i=0;i<1023;++i)
		{
			if(read(clientfd,&buffer2[i],1) != 1)
			{
				fprintf(stderr,"Error: Error reading headers, dumping\n");
				return -1;
			}
			else
			{
				if((buffer2[i] == '\r') && (flag == 1) && (i == 0))	/* this is if we already got one \r\n */
				{
					flag=2;
					break;
				}
				else if(buffer2[i] == '\r')
				{
					flag=0;
					break;
				}
			}
		}

		if((method_num == 2) && (i > 0))			// Just put the "Content-Length:" in buffer to be checked if it really is
		{
			for(j=0;(j < 15) && (j < i); ++j)
			{
				buffer[j] = tolower(buffer2[j]);	// Just make it uppercase so easy to compare
									// Stupid IE does it "Content-Length" while all good browsers
									// do it with "Content-length", which is how the RFC says it!!!
			}
			buffer[j] = 0;		// NULL

			if(strcmp(buffer,"content-length:") == 0)	// Check if the header is indeed content length
			{
				// Read up till 10 chars (content-length can't be more than that, hehe
				for(k=16; (k < i) && (k < (16+10)) && (buffer2[k] != '\r'); ++k)
				{
					content_length[k-16] = buffer2[k];
				}
				content_length[k-16] = 0;		// NULL
			}
		}

		if(read(clientfd,&buffer2[i+1],1) != 1)
		{
			fprintf(stderr,"Error: Error reading headers, dumping\n");
			return -1;
		}

		if((buffer2[i+1] == '\n') && (flag==2))	/* 3 means we got \r\n\r\n and we'd better stop reading headers */
		{
			flag=3;
		}
		else if(buffer2[i+1] == '\n')
		{
			flag=1;
		}
		
		if(write(requestfd,&buffer2,i+2) != (i+2))
		{
			fprintf(stderr,"Error: We couldn't finish writing the headers for client %s, dumping\n", hptr->h_name);
			return -1;
		}
	}

	if(method_num == 2)	// We do special stuff for POST
	{
		if(atoi(content_length) < 1)
		{
			fprintf(stderr, "Error: Content-length seems to be less than 1 byte, this can't be...dumping\n");
			return -1;
		}
		for(i=0; i < atoi(content_length); i++)
		{
			if(read(clientfd,&sendbuffer[0],1) == 1)	// Why waste another variable?  Use first char in array...
			{
				if(write(requestfd,&sendbuffer[0],1) != 1)
				{
					fprintf(stderr,"Error: Sending the POST data to %s from %s...dumping\n",chptr->h_name, hptr->h_name);
					return -1;
				}
			}
		}
//		printf("POST DEBUG: Wrote %d bytes out of what should have been %d bytes\n",i, atoi(content_length));
	}

	// Now we just send all the data verbatim in 512-byte chunks
	while((n = read(requestfd,&sendbuffer,512)) > 0)
	{
		if(write(clientfd,&sendbuffer,n) < 1)
		{
			fprintf(stderr,"Error: Bad write to client %s, dumping!\n", hptr->h_name);
			return -1;
		}
	}
	
	return 0;
}
